"""Base server functionality for LLMling."""

from __future__ import annotations

from typing import Any, cast

from mcp.server import Server as MCPServer
from mcp.server.session import ServerSession
from mcp.shared.context import RequestContext
from mcp.types import (
    CallToolResult,
    EmbeddedResource,
    EmptyResult,
    ImageContent,
    LoggingLevel,
    ServerResult,
    TextContent,
)
from pydantic import AnyUrl

from llmling.core import exceptions
from llmling.core.log import get_logger


logger = get_logger(__name__)


class ServerBase:
    """Base class for LLMling MCP server implementation."""

    def __init__(self, name: str) -> None:
        """Initialize base server.

        Args:
            name: Server name
        """
        self.name = name
        self.server = MCPServer(name)

    @property
    def request_context(self) -> RequestContext[ServerSession]:
        """Get current request context.

        Returns:
            Current request context

        Raises:
            RuntimeError: If called outside request context
        """
        try:
            return cast(RequestContext[ServerSession], self.server.request_context)
        except LookupError as exc:
            msg = "No active request context"
            raise RuntimeError(msg) from exc

    async def notify_progress(
        self,
        progress: float,
        total: float | None = None,
        *,
        description: str | None = None,
    ) -> None:
        """Send progress notification if progress token exists."""
        try:
            ctx = self.request_context
            if ctx.meta and (progress_token := ctx.meta.progressToken):
                # Send progress notification
                await ctx.session.send_progress_notification(
                    progress_token=progress_token,  # Send token first
                    progress=progress,
                    total=total,
                )
                # Send description separately if provided
                if description:
                    await ctx.session.send_log_message(
                        level="info",
                        data=description,
                    )
        except Exception:  # noqa: BLE001
            logger.warning("Failed to send progress notification", exc_info=True)

    async def convert_to_mcp_content(  # noqa: PLR0911
        self,
        content: Any,
        *,
        error: bool = False,
    ) -> list[TextContent | ImageContent | EmbeddedResource]:
        """Convert content to MCP content types.

        Args:
            content: Content to convert
            error: Whether this is error content

        Returns:
            List of MCP content types
        """
        # Handle None
        if content is None:
            return [TextContent(type="text", text="")]

        # Handle strings
        if isinstance(content, str):
            return [TextContent(type="text", text=content)]

        # Handle bytes (convert to base64)
        if isinstance(content, bytes):
            import base64

            data = base64.b64encode(content).decode()
            return [
                TextContent(
                    type="text", text=f"Binary data ({len(content)} bytes): {data[:100]}"
                )
            ]

        # Handle lists and tuples
        if isinstance(content, list | tuple):
            results = []
            for item in content:
                results.extend(await self.convert_to_mcp_content(item))
            return results

        # Handle dictionaries
        if isinstance(content, dict):
            import json

            return [TextContent(type="text", text=json.dumps(content, indent=2))]

        # Handle exceptions
        if isinstance(content, Exception):
            return [TextContent(type="text", text=f"Error: {content!s}")]

        # Handle everything else
        return [TextContent(type="text", text=str(content))]

    async def wrap_result(
        self,
        result: Any,
        *,
        error: bool = False,
    ) -> ServerResult:
        """Wrap result in proper MCP response type."""
        if isinstance(result, ServerResult):
            return result

        try:
            # Convert to MCP content types
            content = await self.convert_to_mcp_content(result, error=error)

            # Get the current request context
            _ctx = self.request_context
            # If this is a tool call result
            if isinstance(content, list) and content:
                return ServerResult(
                    root=CallToolResult(
                        content=content,
                        isError=error,
                    )
                )

            # For empty or non-list results
            return ServerResult(root=EmptyResult())

        except Exception as exc:
            logger.exception("Failed to wrap result")
            msg = f"Failed to process result: {exc}"
            raise exceptions.ProcessorError(msg) from exc

    async def log_message(
        self,
        level: LoggingLevel,
        message: str,
        *,
        logger_name: str | None = None,
    ) -> None:
        """Send log message to client.

        Args:
            level: Log level
            message: Log message
            logger_name: Optional logger name
        """
        try:
            await self.request_context.session.send_log_message(
                level=level,
                data=message,
                logger=logger_name,
            )
        except Exception:  # noqa: BLE001
            logger.warning("Failed to send log message", exc_info=True)

    async def notify_resource_change(self, uri: str) -> None:
        """Notify clients about resource changes."""
        try:
            # Convert string URI to AnyUrl
            url = AnyUrl(uri)
            await self.request_context.session.send_resource_updated(url)
            await self.request_context.session.send_resource_list_changed()
        except Exception:  # noqa: BLE001
            logger.warning("Failed to send resource change notification", exc_info=True)

    async def notify_prompt_change(self) -> None:
        """Notify clients about prompt changes."""
        try:
            await self.request_context.session.send_prompt_list_changed()
        except Exception:  # noqa: BLE001
            logger.warning("Failed to send prompt change notification", exc_info=True)

    async def notify_tool_change(self) -> None:
        """Notify clients about tool changes."""
        try:
            await self.request_context.session.send_tool_list_changed()
        except Exception:  # noqa: BLE001
            logger.warning("Failed to send tool change notification", exc_info=True)



"""Path context loader implementation."""

from __future__ import annotations

from typing import TYPE_CHECKING, ClassVar

from upath import UPath

from llmling.config.models import PathResource
from llmling.core import exceptions
from llmling.core.log import get_logger
from llmling.resources.base import ResourceLoader, create_loaded_resource


if TYPE_CHECKING:
    from llmling.processors.registry import ProcessorRegistry
    from llmling.resources.models import LoadedResource


logger = get_logger(__name__)


class PathResourceLoader(ResourceLoader[PathResource]):
    """Loads context from files or URLs."""

    context_class = PathResource
    uri_scheme = "file"
    supported_mime_types: ClassVar[list[str]] = [
        "text/plain",
        "application/json",
        "text/markdown",
        "text/yaml",
    ]

    @classmethod
    def get_uri_template(cls) -> str:
        """File URIs need three slashes for absolute paths."""
        return "file:///{name}"

    @classmethod
    def create_uri(cls, *, name: str) -> str:
        """Handle file paths properly."""
        normalized = name.replace("\\", "/").lstrip("/")
        return cls.get_uri_template().format(name=normalized)

    async def _load_impl(
        self,
        resource: PathResource,
        name: str,
        processor_registry: ProcessorRegistry | None,
    ) -> LoadedResource:
        """Load content from a file or URL."""
        try:
            path = UPath(resource.path)
            content = path.read_text("utf-8")

            if processor_registry and (procs := resource.processors):
                processed = await processor_registry.process(content, procs)
                content = processed.content

            return create_loaded_resource(
                content=content,
                source_type="path",
                uri=self.create_uri(name=name),
                mime_type=self.supported_mime_types[0],
                name=resource.description or path.name,
                description=resource.description,
                additional_metadata={
                    "type": "path",
                    "path": str(path),
                    "scheme": path.protocol,
                },
            )
        except Exception as exc:
            msg = f"Failed to load content from {resource.path}"
            raise exceptions.LoaderError(msg) from exc


if __name__ == "__main__":
    uri = PathResourceLoader.create_uri(name="/path/to/file.txt")
    print(uri)  # file:///path/to/file.txt



from __future__ import annotations

import asyncio
from unittest.mock import Mock

import pytest

from llmling.config.models import TextResource
from llmling.server.observers import PromptObserver, ResourceObserver, ToolObserver


@pytest.fixture
def mock_server() -> Mock:
    """Create a mock server with required methods."""
    server = Mock()

    # Add async notification methods
    async def notify_change(uri: str) -> None: ...
    async def notify_list_changed() -> None: ...

    server.notify_resource_change = Mock(side_effect=notify_change)
    server.notify_resource_list_changed = Mock(side_effect=notify_list_changed)
    server.notify_prompt_list_changed = Mock(side_effect=notify_list_changed)
    server.notify_tool_list_changed = Mock(side_effect=notify_list_changed)
    # Mock loader registry
    server.loader_registry.get_loader.return_value.create_uri.return_value = "test://uri"
    return server


@pytest.mark.asyncio
async def test_resource_observer_notifications(mock_server: Mock) -> None:
    """Test that resource observer creates notification tasks."""
    observer = ResourceObserver(mock_server)
    resource = TextResource(content="test")

    # Trigger events
    observer._handle_resource_changed("test_key", resource)
    observer._handle_list_changed()

    # Wait for tasks to complete
    await asyncio.sleep(0)

    # Check notifications were triggered
    mock_server.notify_resource_change.assert_called_once_with("test://uri")
    mock_server.notify_resource_list_changed.assert_called_once()


@pytest.mark.asyncio
async def test_prompt_observer_notifications(mock_server: Mock) -> None:
    """Test that prompt observer creates notification tasks."""
    observer = PromptObserver(mock_server)

    # Trigger event
    observer._handle_list_changed()

    # Wait for tasks to complete
    await asyncio.sleep(0)

    # Check notification was triggered
    mock_server.notify_prompt_list_changed.assert_called_once()


@pytest.mark.asyncio
async def test_tool_observer_notifications(mock_server: Mock) -> None:
    """Test that tool observer creates notification tasks."""
    observer = ToolObserver(mock_server)

    # Trigger event
    observer._handle_list_changed()

    # Wait for tasks to complete
    await asyncio.sleep(0)

    # Check notification was triggered
    mock_server.notify_tool_list_changed.assert_called_once()


@pytest.mark.asyncio
async def test_observer_task_cleanup() -> None:
    """Test that observer tasks are properly tracked and cleaned up."""
    server = Mock()

    # Make notification hang for a bit
    async def slow_notify(*args: object) -> None:
        await asyncio.sleep(0.1)

    server.notify_resource_list_changed = slow_notify

    observer = ResourceObserver(server)

    # Create a notification task
    observer._handle_list_changed()

    # Should have one task
    assert len(observer._tasks) == 1

    # Wait for task to complete
    await asyncio.sleep(0.2)

    # Task should be removed
    assert len(observer._tasks) == 0


@pytest.mark.asyncio
async def test_prompt_observer_task_cleanup() -> None:
    """Test that prompt observer tasks are properly tracked and cleaned up."""
    server = Mock()

    async def slow_notify(*args: object) -> None:
        await asyncio.sleep(0.1)

    server.notify_prompt_list_changed = slow_notify

    observer = PromptObserver(server)
    observer._handle_list_changed()
    assert len(observer._tasks) == 1
    await observer.cleanup()
    assert len(observer._tasks) == 0


@pytest.mark.asyncio
async def test_tool_observer_task_cleanup() -> None:
    """Test that tool observer tasks are properly tracked and cleaned up."""
    server = Mock()

    async def slow_notify(*args: object) -> None:
        await asyncio.sleep(0.1)

    server.notify_tool_list_changed = slow_notify

    observer = ToolObserver(server)
    observer._handle_list_changed()
    assert len(observer._tasks) == 1
    await observer.cleanup()
    assert len(observer._tasks) == 0
